library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;
use work.quarkpkg.all;

entity controller is

  port (
    ClkxCI    : in  std_logic;
    RstxRBI   : in  std_logic;
    StartxSI  : in  std_logic;
    INENxEI   : in  std_logic;
    FreezexSO : out std_logic;
    OUTENxEO  : out std_logic);

end controller;

architecture rtl of controller is
  
  type state is (idle, update, final);
  signal StatexDP, StatexDN : state;

  signal RcntxDP, RcntxDN : unsigned(9 downto 0);
  signal FcntxDP, FcntxDN : unsigned(4 downto 0);
  
begin  -- rtl

  p_fsm: process (FcntxDP, INENxEI, RcntxDP, StartxSI, StatexDP)
  begin  -- process p_fsm

    StatexDN  <= StatexDP;
    RcntxDN   <= (others => '0');
    FcntxDN   <= (others => '0');
    OUTENxEO  <= '0';
    FreezexSO <= '0';

    case StatexDP is
      -------------------------------------------------------------------------
      when idle =>

        if StartxSI = '1' and INENxEI = '1' then
          StatexDN <= update;

        else
          FreezexSO <= '1';
          
        end if;

      -------------------------------------------------------------------------
      when update =>

        if RcntxDP = ROUND-1 then
          if INENxEI = '0' then
            OUTENxEO <= '1';
            StatexDN <= final;
          
          end if;

        else
          RcntxDN <= RcntxDP + 1;
          
        end if;

      -------------------------------------------------------------------------
      when final =>

        if RcntxDP = ROUND-1 then   
          OUTENxEO <= '1';
            
          if FcntxDP = DIGEST/RATE-1 then
            StatexDN <= idle;

          else
            FcntxDN <= FcntxDP + 1;
            
          end if;

        else
          FcntxDN <= FcntxDP;
          RcntxDN <= RcntxDP + 1;
          
        end if;

      -------------------------------------------------------------------------
      when others => StatexDN <= idle;
                     
    end case;
    
  end process p_fsm;

  p_mem : process (ClkxCI, RstxRBI)
  begin  -- process p_mem
    if RstxRBI = '0' then               -- asynchronous reset (active low)
      StatexDP <= idle;
      RcntxDP  <= (others => '0');
      FcntxDP  <= (others => '0');
      
    elsif ClkxCI'event and ClkxCI = '1' then  -- rising clock edge
      StatexDP <= StatexDN;
      RcntxDP  <= RcntxDN;
      FcntxDP  <= FcntxDN;
      
    end if;
  end process p_mem;

end rtl;
