library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;
use work.quarkpkg.all;

entity quark is

  port (
    ClkxCI   : in  std_logic;
    RstxRBI  : in  std_logic;
    StartxSI : in  std_logic;
    INENxEI  : in  std_logic;
    OUTENxEO : out std_logic;
    DxDI     : in  std_logic_vector(RATE*8-1 downto 0);
    DxDO     : out std_logic_vector(RATE*8-1 downto 0));

end quark;

architecture rtl of quark is

  component controller
    port (
      ClkxCI    : in  std_logic;
      RstxRBI   : in  std_logic;
      StartxSI  : in  std_logic;
      INENxEI   : in  std_logic;
      FreezexSO : out std_logic;
      OUTENxEO  : out std_logic);
  end component;

  signal QuarkStatexDP, QuarkStatexDN : std_logic_vector(0 to WWIDTH*8-1);
  signal PermOutxD                    : std_logic_vector(0 to WWIDTH*8-1);
  signal X, Y                         : std_logic_vector(0 to WWIDTH*8/2-1);
  signal LxDN, LxDP, LoutxD           : std_logic_vector(0 to LWIDTH-1);
  signal Xn, Yn, Ln, h, Xnn, Ynn      : std_logic_vector(0 to PDEG-1);
  signal OUTENxE, FreezexS            : std_logic;
  signal DInxD                        : std_logic_vector(0 to RATE*8-1);
  signal IR, IW, ID                   : std_logic_vector(7 downto 0);
  
begin  -- rtl
  
  f_inout : for i in 0 to RATE-1 generate
    f_inout2 : for j in 0 to 7 generate
      DInxD(8*i+j)           <= DxDI(RATE*8-8*i-(8-j));
      DxDO(RATE*8-8*i-(j+1)) <= PermOutxD(8*(WWIDTH-RATE+i)+j);
    end generate f_inout2;
  end generate f_inout;

  OUTENxEO <= OUTENxE;

  -- FSM
  -----------------------------------------------------------------------------
  u_controller : controller
    port map (
      ClkxCI    => ClkxCI,
      RstxRBI   => RstxRBI,
      StartxSI  => StartxSI,
      INENxEI   => INENxEI,
      FreezexSO => FreezexS,
      OUTENxEO  => OUTENxE);

  -- Input Select
  -----------------------------------------------------------------------------
  QuarkStatexDN(0 to 8*(WWIDTH-RATE)-1) <= INITSTATE(0 to 8*(WWIDTH-RATE)-1) when StartxSI = '1' and INENxEI = '1' else
                                           PermOutxD(0 to 8*(WWIDTH-RATE)-1);

  QuarkStatexDN(8*(WWIDTH-RATE) to 8*WWIDTH-1) <= INITSTATE(8*(WWIDTH-RATE) to 8*WWIDTH-1) xor DinxD when StartxSI = '1' and INENxEI = '1' else
                                                  PermOutxD(8*(WWIDTH-RATE) to 8*WWIDTH-1) xor DinxD when StartxSI = '0' and INENxEI = '1' else
                                                  PermOutxD(8*(WWIDTH-RATE) to 8*WWIDTH-1);

  
  LxDN <= (others => '1') when INENxEI = '1' or StartxSI = '1' or OUTENxE = '1' else LoutxD;

  -- Permutation
  -----------------------------------------------------------------------------

  X <= QuarkStatexDP(0 to WWIDTH*8/2-1);
  Y <= QuarkStatexDP(WWIDTH*8/2 to WWIDTH*8-1);

  perm : for i in 0 to PDEG-1 generate

    Xn(i) <= Y(0+i) xor X(0+i) xor X(11+i) xor X(18+i) xor X(27+i) xor X(36+i) xor
             X(i+42) xor X(i+47) xor X(i+58) xor X(i+67) xor X(i+71) xor X(i+64) xor
             (X(i+79) and X(i+71)) xor (X(i+47) and X(i+42)) xor (X(i+19) and X(i+11)) xor
             (X(i+71) and X(i+67) and X(i+58)) xor (X(i+42) and X(i+36) and X(i+27)) xor
             (X(i+79) and X(i+58) and X(i+36) and X(i+11)) xor
             (X(i+71) and X(i+67) and X(i+47) and X(i+42)) xor
             (X(i+79) and X(i+71) and X(i+27) and X(i+19)) xor
             (X(i+79) and X(i+71) and X(i+67) and X(i+58) and X(i+47)) xor
             (X(i+42) and X(i+36) and X(i+27) and X(i+19) and X(i+11)) xor
             (X(i+67) and X(i+58) and X(i+47) and X(i+42) and X(i+36) and X(i+27));
    
    Yn(i) <= Y(0+i) xor Y(9+i) xor Y(20+i) xor Y(25+i) xor Y(38+i) xor
             Y(i+44) xor Y(i+47) xor Y(i+54) xor Y(i+67) xor Y(i+69) xor Y(i+63) xor
             (Y(i+78) and Y(i+69)) xor (Y(i+47) and Y(i+44)) xor (Y(i+19) and Y(i+9)) xor
             (Y(i+69) and Y(i+67) and Y(i+54)) xor (Y(i+44) and Y(i+38) and Y(i+25)) xor
             (Y(i+78) and Y(i+54) and Y(i+38) and Y(i+9)) xor
             (Y(i+69) and Y(i+67) and Y(i+47) and Y(i+44)) xor
             (Y(i+78) and Y(i+69) and Y(i+25) and Y(i+19)) xor
             (Y(i+78) and Y(i+69) and Y(i+67) and Y(i+54) and Y(i+47)) xor
             (Y(i+44) and Y(i+38) and Y(i+25) and Y(i+19) and Y(i+9)) xor
             (Y(i+67) and Y(i+54) and Y(i+47) and Y(i+44) and Y(i+38) and Y(i+25));

    h(i) <= LxDP(i) xor X(1+i) xor Y(2+i) xor X(5+i) xor Y(12+i) xor Y(24+i) xor X(35+i) xor X(40+i) xor X(48+i) xor Y(55+i) xor
            Y(61+i) xor X(72+i) xor Y(79+i) xor (Y(4+i) and X(68+i)) xor (X(57+i) and X(68+i)) xor (X(68+i) and Y(79+i)) xor
            (Y(4+i) and X(35+i) and X(57+i)) xor (Y(4+i) and X(57+i) and X(68+i)) xor (Y(4+i) and X(57+i) and Y(79+i)) xor
            (LxDP(i) and X(35+i) and X(57+i) and Y(79+i)) xor (LxDP(i) and X(35+i));
    
    Xnn(i) <= Xn(i) xor h(i);
    Ynn(i) <= Yn(i) xor h(i);
    
  end generate perm;

  Ln(0) <= LxDP(0) xor LxDP(3);
  Ln(1) <= LxDP(1) xor LxDP(4);
  Ln(2) <= LxDP(2) xor LxDP(5);
  Ln(3) <= LxDP(3) xor LxDP(6);
  Ln(4) <= LxDP(4) xor LxDP(7);
  Ln(5) <= LxDP(5) xor LxDP(8);
  Ln(6) <= LxDP(6) xor LxDP(9);
  Ln(7) <= LxDP(7) xor Ln(0);


  PermOutxD <= QuarkStatexDP(PDEG to WWIDTH*8/2-1) & Xnn & QuarkStatexDP(WWIDTH*8/2+PDEG to WWIDTH*8-1) & Ynn;
  LoutxD    <= LxDP(PDEG to LWIDTH-1) & Ln;

  -- State Register
  -----------------------------------------------------------------------------
  p_mem : process (ClkxCI, RstxRBI)
  begin  -- process p_mem
    if RstxRBI = '0' then               -- asynchronous reset (active low)
      QuarkStatexDP <= (others => '0');
      LxDP          <= (others => '0');
      
    elsif ClkxCI'event and ClkxCI = '1' then  -- rising clock edge
      --if FreezexS = '0' then
      QuarkStatexDP <= QuarkStatexDN;
      LxDP          <= LxDN;

      --end if;
      
    end if;
  end process p_mem;

end rtl;
