library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;
use work.quarkpkg.all;

entity quark is

  port (
    ClkxCI   : in  std_logic;
    RstxRBI  : in  std_logic;
    StartxSI : in  std_logic;
    INENxEI  : in  std_logic;
    OUTENxEO : out std_logic;
    DxDI     : in  std_logic_vector(RATE*8-1 downto 0);
    DxDO     : out std_logic_vector(RATE*8-1 downto 0));

end quark;

architecture rtl of quark is

  component controller
    port (
      ClkxCI    : in  std_logic;
      RstxRBI   : in  std_logic;
      StartxSI  : in  std_logic;
      INENxEI   : in  std_logic;
      FreezexSO : out std_logic;
      OUTENxEO  : out std_logic);
  end component;

  signal QuarkStatexDP, QuarkStatexDN : std_logic_vector(0 to WWIDTH*8-1);
  signal PermOutxD                    : std_logic_vector(0 to WWIDTH*8-1);
  signal X, Y                         : std_logic_vector(0 to WWIDTH*8/2-1);
  signal LxDN, LxDP, LoutxD           : std_logic_vector(0 to LWIDTH-1);
  signal Xn, Yn, Ln, h, Xnn, Ynn      : std_logic_vector(0 to PDEG-1);
  signal OUTENxE, FreezexS            : std_logic;
  signal DInxD                        : std_logic_vector(0 to RATE*8-1);
  signal IR, IW, ID                   : std_logic_vector(7 downto 0);
  
begin  -- rtl
  
  f_inout : for i in 0 to RATE-1 generate
    f_inout2 : for j in 0 to 7 generate
      DInxD(8*i+j)           <= DxDI(RATE*8-8*i-(8-j));
      DxDO(RATE*8-8*i-(j+1)) <= PermOutxD(8*(WWIDTH-RATE+i)+j);
    end generate f_inout2;
  end generate f_inout;

  OUTENxEO <= OUTENxE;

  -- FSM
  -----------------------------------------------------------------------------
  u_controller : controller
    port map (
      ClkxCI    => ClkxCI,
      RstxRBI   => RstxRBI,
      StartxSI  => StartxSI,
      INENxEI   => INENxEI,
      FreezexSO => FreezexS,
      OUTENxEO  => OUTENxE);

  -- Input Select
  -----------------------------------------------------------------------------
  QuarkStatexDN(0 to 8*(WWIDTH-RATE)-1) <= INITSTATE(0 to 8*(WWIDTH-RATE)-1) when StartxSI = '1' and INENxEI = '1' else
                                           PermOutxD(0 to 8*(WWIDTH-RATE)-1);

  QuarkStatexDN(8*(WWIDTH-RATE) to 8*WWIDTH-1) <= INITSTATE(8*(WWIDTH-RATE) to 8*WWIDTH-1) xor DinxD when StartxSI = '1' and INENxEI = '1' else
                                                  PermOutxD(8*(WWIDTH-RATE) to 8*WWIDTH-1) xor DinxD when StartxSI = '0' and INENxEI = '1' else
                                                  PermOutxD(8*(WWIDTH-RATE) to 8*WWIDTH-1);
  
  LxDN <= (others => '1') when INENxEI = '1' or StartxSI = '1' or OUTENxE = '1' else LoutxD;

  -- Permutation
  -----------------------------------------------------------------------------

  X <= QuarkStatexDP(0 to WWIDTH*8/2-1);
  Y <= QuarkStatexDP(WWIDTH*8/2 to WWIDTH*8-1);

  perm : for i in 0 to PDEG-1 generate

    Xn(i) <= Y(i) xor X(i) xor X(16+i) xor X(26+i) xor X(39+i) xor X(52+i) xor
             X(61+i) xor X(69+i) xor X(84+i) xor X(97+i) xor X(103+i) xor X(94+i) xor
             (X(111+i) and X(103+i)) xor (X(69+i) and X(61+i)) xor (X(28+i) and X(16+i)) xor
             (X(103+i) and X(97+i) and X(84+i)) xor (X(61+i) and X(52+i) and X(39+i)) xor
             (X(111+i) and X(84+i) and X(52+i) and X(16+i)) xor
             (X(103+i) and X(97+i) and X(69+i) and X(61+i)) xor
             (X(111+i) and X(103+i) and X(39+i) and X(28+i)) xor
             (X(111+i) and X(103+i) and X(97+i) and X(84+i) and X(69+i)) xor
             (X(61+i) and X(52+i) and X(39+i) and X(28+i) and X(16+i)) xor
             (X(97+i) and X(84+i) and X(69+i) and X(61+i) and X(52+i) and X(39+i));
    
    Yn(i) <= Y(i) xor Y(13+i) xor Y(30+i) xor Y(37+i) xor Y(56+i) xor
             Y(65+i) xor Y(69+i) xor Y(79+i) xor Y(96+i) xor Y(101+i) xor Y(92+i) xor
             (Y(109+i) and Y(101+i)) xor (Y(69+i) and Y(65+i)) xor (Y(28+i) and Y(13+i)) xor
             (Y(101+i) and Y(96+i) and Y(79+i)) xor (Y(65+i) and Y(56+i) and Y(37+i)) xor
             (Y(109+i) and Y(79+i) and Y(56+i) and Y(13+i)) xor
             (Y(101+i) and Y(96+i) and Y(69+i) and Y(65+i)) xor
             (Y(109+i) and Y(101+i) and Y(37+i) and Y(28+i)) xor
             (Y(109+i) and Y(101+i) and Y(96+i) and Y(79+i) and Y(69+i)) xor
             (Y(65+i) and Y(56+i) and Y(37+i) and Y(28+i) and Y(13+i)) xor
             (Y(96+i) and Y(79+i) and Y(69+i) and Y(65+i) and Y(56+i) and Y(37+i));
    
    Xnn(i) <= Xn(i) xor h(i);
    Ynn(i) <= Yn(i) xor h(i);
    
  end generate perm;

  perm_h10 : for i in 0 to 9 generate
    
    h(i) <= LxDP(i) xor X(1+i) xor Y(3+i) xor X(7+i) xor Y(18+i) xor Y(34+i) xor X(47+i) xor X(58+i) xor Y(71+i) xor Y(80+i) xor
            X(90+i) xor Y(91+i) xor X(105+i) xor Y(111+i) xor (Y(8+i) and X(100+i)) xor (X(72+i) and X(100+i)) xor (X(100+i) and Y(111+i)) xor
            (Y(8+i) and X(47+i) and X(72+i)) xor (Y(8+i) and X(72+i) and X(100+i)) xor (Y(8+i) and X(72+i) and Y(111+i)) xor
            (LxDP(i) and X(47+i) and X(72+i) and Y(111+i)) xor (LxDP(i) and X(47+i));

  end generate perm_h10;

  perm_h16 : for i in 10 to PDEG-1 generate
    
    h(i) <= Ln(i-10) xor X(1+i) xor Y(3+i) xor X(7+i) xor Y(18+i) xor Y(34+i) xor X(47+i) xor X(58+i) xor Y(71+i) xor Y(80+i) xor
            X(90+i) xor Y(91+i) xor X(105+i) xor Y(111+i) xor (Y(8+i) and X(100+i)) xor (X(72+i) and X(100+i)) xor (X(100+i) and Y(111+i)) xor
            (Y(8+i) and X(47+i) and X(72+i)) xor (Y(8+i) and X(72+i) and X(100+i)) xor (Y(8+i) and X(72+i) and Y(111+i)) xor
            (Ln(i-10) and X(47+i) and X(72+i) and Y(111+i)) xor (Ln(i-10) and X(47+i));

  end generate perm_h16;

  Ln(0) <= LxDP(0) xor LxDP(3);
  Ln(1) <= LxDP(1) xor LxDP(4);
  Ln(2) <= LxDP(2) xor LxDP(5);
  Ln(3) <= LxDP(3) xor LxDP(6);
  Ln(4) <= LxDP(4) xor LxDP(7);
  Ln(5) <= LxDP(5) xor LxDP(8);
  Ln(6) <= LxDP(6) xor LxDP(9);

  Ln(7) <= LxDP(7) xor Ln(0);
  Ln(8) <= LxDP(8) xor Ln(1);
  Ln(9) <= LxDP(9) xor Ln(2);

  Ln(10) <= LxDP(0) xor LxDP(6);
  Ln(11) <= LxDP(1) xor LxDP(7);
  Ln(12) <= LxDP(2) xor LxDP(8);
  Ln(13) <= LxDP(3) xor LxDP(9);

  Ln(14) <= LxDP(4) xor Ln(0);
  Ln(15) <= LxDP(5) xor Ln(1);


  PermOutxD <= QuarkStatexDP(PDEG to WWIDTH*8/2-1) & Xnn & QuarkStatexDP(WWIDTH*8/2+PDEG to WWIDTH*8-1) & Ynn;
  LoutxD    <= Ln(PDEG-10 to PDEG-1);

  -- State Register
  -----------------------------------------------------------------------------
  p_mem : process (ClkxCI, RstxRBI)
  begin  -- process p_mem
    if RstxRBI = '0' then               -- asynchronous reset (active low)
      QuarkStatexDP <= (others => '0');
      LxDP          <= (others => '0');
      
    elsif ClkxCI'event and ClkxCI = '1' then  -- rising clock edge
      --if FreezexS = '0' then
      QuarkStatexDP <= QuarkStatexDN;
      LxDP          <= LxDN;

      --end if;
      
    end if;
  end process p_mem;

end rtl;
